/* 
 * **  KeenView -- Pseudo-tty interface para linha multiponto  **  (C)
 * 1997 -- Rildo Pragana <rpragana@acm.org> 
 * $Id: kvpty.c,v 1.4 1998/08/11 23:36:59 root Exp $ 
 * $Log: kvpty.c,v $
 * Revision 1.4  1998/08/11 23:36:59  root
 * removido bug na saida (exit) das estacoes
 *
 */

#include 	<stdio.h>
#include 	<stdlib.h>
#include 	<signal.h>
#include 	<sys/time.h>
#include 	<sys/types.h>
#include 	<sys/ioctl.h>
#include 	<sys/wait.h>
#include 	<termios.h>
#include 	<string.h> 
#include 	<fcntl.h>
#include	<errno.h>
#include	<pwd.h>
#include	<unistd.h>

#include	"kvpty.h"

#define PRTBUSY 2
#define isechoed(c) ((c >= ' ') || (c == '\n') || (c == ' '))
#define max(a,b) (((a) > (b)) ? (a) : (b))
#define min(a,b) (((a) < (b)) ? (a) : (b))

char *kvcommand_default = "/bin/ash";
char *kvcommands[] =
{NULL, NULL};

struct termios stbuf, svbuf;
char *device = "/dev/kvadap";

#ifdef NEED_SELECT_H
#include	<sys/select.h>
#endif

#ifdef NEED_INET_H
#include	<sys/inet.h>
#endif

#ifndef SIGCLD	
#define SIGCLD SIGCHLD
#endif

/*
 * ptys e pids das estacoes 
 */
int kvpresente[32];
int kvcol[32], kvlin[32];
struct kvwin {
  int scol, slin;
  int stat;
  char scr[25][80];
} kvwin[32];
int kvpid[32];

int kvptyfd[32];	/* interface com os pty's */
int kvdevfd[32];	/* interface com os dispositivos */
int prtpid[32];

/*************************************
 Atualizao do display dos terminais
*************************************/
void
kv_update (int n, int col, int lin)
{
	unsigned char c;
	int i, j;

#ifdef SHOW_UPDATE
	printf("\nkv_update:\n1==>");
#endif
	write( kvdevfd[n],"\x1\x0\x0\x0\x0\x0\x0\x0\x0",9 );
	for (i = col; i < col + 16; i++) {
		c = kvwin[n].scr[lin][i];
#ifdef SHOW_UPDATE
		if ((c < ' ') || (c > '\x7f')) {
			printf("[[%02x]]",c);
		} else {
#endif
			write( kvdevfd[n],&c,1 );
#ifdef SHOW_UPDATE
			putchar( c );
		}
#endif
	}
#ifdef SHOW_UPDATE
	printf("\n2==>");
#endif
	write( kvdevfd[n],"\xc0\x0\x0\x0\x0\x0\x0\x0\x0",9 );
	for (i = col; i < col + 16; i++) {
		c = kvwin[n].scr[lin+1][i];
#ifdef SHOW_UPDATE
		if ((c < ' ') || (c > '\x7f')) {
			printf("[[%02x]]",c);
		} else {
#endif
			write( kvdevfd[n],&c,1 );
#ifdef SHOW_UPDATE
			putchar( c );
		}
#endif
	}
	c = ((kvlin[n] - kvwin[n].slin)&1) ? 0xc0 : 0x80;
	c += (kvcol[n] - kvwin[n].scol);
	write( kvdevfd[n],&c,1 );
#ifdef SHOW_UPDATE
	printf("\ncursor(%02x)-------------------",c);
#endif
}

void 
init_screen ( int n )
{
	kvwin[n].scol = kvwin[n].slin = 0;
	kvwin[n].stat = 0;
	kvcol[n] = 0;
	kvlin[n] = 1;
}

void
kv_home (int n)
{
	int i,j;
	write( kvdevfd[n],"\x1\x0\x0\x0\x0\x0\x0\x0\x0",9 );
	for (j = 0; j < 24; j++)
 		for (i = 0; i < 80; i++)
    		kvwin[n].scr[j][i] = ' ';
    kvwin[n].slin = kvwin[n].scol = 0;
}

/******************************************
 movimento da janela visvel nos terminais
******************************************/
void
move_win (int ch, int n)
{
  switch (ch) {
  case 1:
    kvwin[n].scol = max (kvwin[n].scol - 8, 0);
    break;
  case 2:
    kvwin[n].scol = min (kvwin[n].scol + 8, (80 - 16));
    break;
  case 3:
    kvwin[n].slin = max (kvwin[n].slin - 1, 0);
    break;
  case 4:
    kvwin[n].slin = min (kvwin[n].slin + 1, 22);
    break;
  case 5:
    kv_home(n);
    break;
  case 6:
    kvwin[n].slin = min ((kvlin[n]), 22);
    kvwin[n].scol = min ((kvcol[n] / 8) * 8, (80 - 16));
    break;
  }
  kv_update (n, kvwin[n].scol, kvwin[n].slin);
}

/*******************************************
 Envia caracter para o display, processando
 caracteres de controle.
*******************************************/
int
kv_write (int n, char *buffer, int len)
{
  int i, j;
  char ch[8];
  int update_needed;
  update_needed = 0;
  while (len--) {
    switch (kvwin[n].stat) {
    case 0:
      switch (*buffer) {
      case '\e':
		kvwin[n].stat = 1;
		break;
      case '\f':
		kvlin[n] = 0;
		kvcol[n] = 0;
/*		for (j = 0; j < 24; j++)
	  		for (i = 0; i < 80; i++)
	    		kvwin[n].scr[j][i] = ' ';
*/
		kv_home(n);
		update_needed |= 1;
		break;
      case '\n':		
		if (kvlin[n] == 23) {
	  		for (j = 1; j < 24; j++)
	    		for (i = 0; i < 80; i++)
	      			kvwin[n].scr[j - 1][i] = kvwin[n].scr[j][i];
	  		for (i = 0; i < 80; i++)
	    		kvwin[n].scr[23][i] = ' ';
		}
		kvcol[n] = 0;
		if (kvlin[n] < 23)
	  		kvlin[n]++;
		update_needed |= 1;
		break;
      case '\r':
		kvcol[n] = 0;
		update_needed |= 1;
		break;
      case '\b':
		if (kvcol[n]) {
	  		kvcol[n]--;
	  		update_needed |= 1;
		}
		break;
      case '\t':
		if (kvcol[n] < 80) {
		  	kvwin[n].scr[kvlin[n]][kvcol[n]] = ' ';
	  		kvcol[n]++;
		}
		break;
      case '\x7':
		break;
      default:
		if (kvcol[n] < 80) {
	  		kvwin[n].scr[kvlin[n]][kvcol[n]] = 
	  			(((*buffer < ' ')||(*buffer < 0)) ? ' ' : *buffer);
	  		kvcol[n]++;
		}
      }
      break;
	/*** tratando sequencia de escape ***/
    case 1:
      switch (*buffer) {
		case 'H': 
			kvcol[n]=0; 
			kvlin[n]=0; 
			kvwin[n].stat=0; 
			break; 
    	case 'R':
			for (j = 1; j < 24; j++)
	  			for (i = 0; i < 80; i++)
	    			kvwin[n].scr[j - 1][i] = kvwin[n].scr[j][i];
			for (i = 0; i < 80; i++)
	  			kvwin[n].scr[23][i] = ' ';
			kvwin[n].stat = 0;
			break;
      	case 'A':
			if (kvlin[n] > 0)
	  			kvlin[n]--;
			kvwin[n].stat = 0;
			break;
      	case 'D':
			if (kvcol[n] > 0)
	  			kvcol[n]--;
			kvwin[n].stat = 0;
			break;
      	case 'C':
			if (kvcol[n] < 79)
		  		kvcol[n]++;
			kvwin[n].stat = 0;
			break;
      	case 'B':
			if (kvlin[n] < 22)
	  			kvlin[n]++;
			kvwin[n].stat = 0;
			break;
      	case 'J':
			for (i = kvcol[n]; i < 80; i++)
	  			kvwin[n].scr[kvlin[n]][i] = ' ';
			for (j = kvlin[n] + 1; j < 24; j++)
	  			for (i = 0; i < 80; i++)
	    			kvwin[n].scr[j][i] = ' ';
			kvwin[n].stat = 0;
			break;
      	case 'K':
			for (i = kvcol[n]; i < 80; i++)
	  			kvwin[n].scr[kvlin[n]][i] = ' ';
			kvwin[n].stat = 0;
			break;
      	case 'Y':
			kvwin[n].stat = 2;
			break;
      	case 'I':
			for (j = 0; j < 22; j++)
	  			for (i = 0; i < 80; i++)
	    			kvwin[n].scr[j + 1][i] = kvwin[n].scr[j][i];
			for (i = 0; i < 80; i++)
	  			kvwin[n].scr[0][i] = ' ';
			kvwin[n].stat = 0;
			break;
	  	case '(': 
	  		kvwin[n].stat=4; 
	  		break;
      	default:
			kvwin[n].stat = 0;
			break;
      }
      break;
    case 2:
      kvlin[n] = *buffer - ' ';
      kvwin[n].stat = 3;
      break;
    case 3:
      kvcol[n] = *buffer - ' ';
      kvwin[n].stat = 0;
      break;
    case 4:
      kvwin[n].stat = 0;
      break;
	default:
		kvwin[n].stat = 0;
		break;
    }
    buffer++;
  }
  if (kvwin[n].stat) return;
  	kvwin[n].scol = kvcol[n] - 16;
  	if (kvwin[n].scol < 0) kvwin[n].scol = 0;
  	if (kvwin[n].scol > (80-16)) kvwin[n].scol = (80-16);
  	kvwin[n].slin = kvlin[n]-1;
  	if (kvwin[n].slin > 22)
    	kvwin[n].slin = 22;
  	if (kvwin[n].slin < 0)
    	kvwin[n].slin = 0;
  	kv_update (n, kvwin[n].scol, kvwin[n].slin);

}

void process_options() {
  	unsigned char tmp[20];
 	char *s, *option, *option_global;
 	unsigned char buffer[BUFSIZ]; 
	int i;
	 
  	/* obtem do environment estacoes a ativar */
  	if ((s = getenv ("KVSTATIONS")) != NULL) {
    	for (i = 0; i < 32; i++)
      		kvpresente[i] = 0;
    	while (*s) {
      		while (isblank (*s))
				s++;
      		i = 0;
      		while (isdigit (*s))
				i = i * 10 + (*s++) - '0';
      		/* aqui criaremos um processo (com pty?)  para cada estacao */
      		kvpresente[i]++;
      		if (*s == ':')
				s++;
      		else
				break;
    	}
  	}
  	else {
    	printf ("Por favor set KVSTATIONS com lista das estaes a serem ativadas!\n");
    	exit (1);
  	}

  	/* obtem estacoes com impressoras conectadas */
  	if ((s = getenv ("KVPRINTERS")) != NULL) {
    while (*s) {
      	while (isblank (*s))
			s++;
      	i = 0;
      	while (isdigit (*s))
			i = i * 10 + (*s++) - '0';
		/* cria um processo gerenciando esta impressora */
		prtpid[i] = mk_prt_manager(i);
      	if (*s == ':')
			s++;
      	else
			break;
    	}
  	}
  	/* obtem do environment kvcommands default */
  	if ((option_global = getenv ("KVCOMMAND")) != NULL)
    	kvcommands[0] = option_global;
  	else
    	kvcommands[0] = option_global = kvcommand_default;
  	for (i = 0; i < 32; i++) {
    /* tenta inicialmente comando especifico por estacao */
    	sprintf (buffer, "KVCOMMAND%d", i);
    	if ((option = getenv (buffer)) != NULL)
      		kvcommands[0] = option;
    	else
    		kvcommands[0] = option_global;
	}
}

void 
sighandler( int sign )
{
	int i,pid;
	for (i=0;i<32;i++) {
		if (prtpid[i])
			kill( prtpid[i],9 );
	}
	/* determine which child has sent us this signal */
	pid = wait(NULL);
	for (i=0;i<32;i++) {
		if (kvpid[i] == pid)
			break;
	}
	printf("sighandler: signal %d, terminal %d\n",sign,i);
	exit(1);
}

/*************************************
**************************************
**									**
**   M   M    A    III    N   N     **
**   MM MM   A A    I     NN  N     **
**   M M M  A   A   I     N N N     **
**   M M M  AAAAA   I     N  NN     **
**   M   M  A   A   I     N   N     **
**   M   M  A   A  III    N   N     **
**                                  **
**************************************
*************************************/

int
main (int argc, char *argv[])
{
  	char *s;
	char buf[BUFSIZ];
  	extern int errno;
  	int i, j, jj;
	char *ptr, c;
  	char *xterm_title = NULL;
  	struct timeval tv;
  	fd_set ready;

  	printf ("\n\nKeenView Interface via PTYs\n");
	signal( SIGABRT, sighandler );
	signal( SIGCHLD, sighandler );

	process_options();
  	for (i = 0; i < 32; i++) {
    	if (kvpresente[i]) {
    		if ((kvptyfd[i] = pty_open (kvcommands, &kvpid[i], i)) < 0) {
				switch (errno) {
				case EIO:
				case EPERM:
				case ENOENT:
	  				fprintf (stderr, "No available pseudo terminals.\n");
	  				break;
				case EAGAIN:
		  			fprintf (stderr, "No more processes, try again later.\n");
		  			break;
				default:
	  				perror ("pty_open() error");
				}
	      	} else {
	      		init_screen(i);
				kvdevfd[i] = open(device,O_RDWR | O_NONBLOCK);
				ioctl( kvdevfd[i],KVADAP_IOCSADDR,i );
	      		kv_home(i);
	      	}
    	}
  	}
	while(1) {
		FD_ZERO(&ready);
		for (i=0;i<32;i++) {
			if (kvpresente[i]) {
				FD_SET(kvptyfd[i],&ready);
				FD_SET(kvdevfd[i],&ready);
			}
		}
		FD_SET( 0,&ready );
		
		select(FD_SETSIZE,&ready, NULL, NULL, NULL);
		if ( FD_ISSET( 0,&ready ) ) {
			read( 0,buf,BUFSIZ );
			ioctl( kvdevfd[0],KVADAP_IOCRESET );
		}
for (i=0;i<32;i++) {
	if (kvpresente[i]) {
		if (FD_ISSET(kvptyfd[i],&ready)) {
			j = read( kvptyfd[i],buf,BUFSIZ );
			kv_write( i,buf,j );
		}
		if (FD_ISSET(kvdevfd[i],&ready)) {
			j = read( kvdevfd[i],buf,1 );
			if (j>0) {
				if ((unsigned char)*buf > 0x90) {
					write( kvptyfd[i],"\x1b",1 );
					write( kvptyfd[i],buf,1 );
				} else if ((unsigned char)*buf > 0x80) {
					move_win( *buf & 0x7f,i );
				} else {
					write( kvptyfd[i],buf,1 );
				}
			}
		}
	}
}	
	} /* while */
  	return (0);
}
