/* 
** Miscellaneous system dependent routines
** taken from splitvt.
** Rildo Pragana, 1998
** $Id: kvptyaux.c,v 1.3 1998/08/11 23:36:59 root Exp $
*/

#include <unistd.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<sys/time.h>
#include	<fcntl.h>
#include	<stdio.h>
#include	<signal.h>
#include	<termios.h>

#include "kvpty.h"

/*
 * create a new process to manage printer fifo
 */
 
int mk_prt_manager( prn ) 
int prn;
{
	char name[80];
	char buf[80];
	int pid,fd,tmp,device,rc;
	fd_set watchset;
	sprintf( name,"/dev/kvprt%d",prn );
	pid = fork();
	if (pid < 0) {
		perror("printer: fork failed!");
		exit(128);
	}
	if (pid == 0) {
		/* prepare physical device for writing */
		device = open("/dev/kvadap",O_WRONLY);
		ioctl(device,KVADAP_IOCSADDR,prn+32);
			
		while(1) {
			if ((fd=open(name,O_RDONLY | O_NONBLOCK)) < 0) {
				perror("printer: open failed");
				exit(1);
			}
			FD_ZERO(&watchset);
			FD_SET(fd,&watchset);
			do {
				if (select(FD_SETSIZE,&watchset,NULL,NULL,NULL) < 0) {
					perror("printer: select failed");
					exit(1);
				}
				rc = read(fd,buf,sizeof(buf) - 1);
				if (rc < 0) {
					perror("printer: read failed");
					exit(1);
				} else {
					buf[rc] = '\0';
					/* send buffer to printer */
					write(device,buf,rc);
				}
			} while(rc);
			close(fd);
		}
	}
	return pid;
}
 
/*
 * Initialize a pty, fork a command running under it, and then 
 * return the master file descriptor
 */

int pty_open(argv, childpid, addr)
char *argv[];
int *childpid;
int addr;		
{

	int get_master_pty(), get_slave_pty();
	char *get_ttyname(), *myputenv();
	int returnfd, slave_fd;
	char envline[128];
	char str_addr[20];

	/* Get the master pty file descriptor */
	if ( (returnfd=get_master_pty()) < 0 )
		return(-1);

	/* Fork and go! */
	if ( ((*childpid)=fork()) < 0 )
		return(-1);
	else if ( (*childpid) == 0 )
	{
		setsid(); /* retira o terminal controlador */
		
		if ( (slave_fd=get_slave_pty()) < 0 )
		{
			perror("Can't open slave tty");
			exit(128);
		}
		close(0); close(1); close(2);
		dup(slave_fd); dup(slave_fd); dup(slave_fd);
		close(slave_fd); close(returnfd);

		/* Reattatch the new tty as the controlling terminal */
		/* Under old UNIX, just opening the new tty after
		   losing the controlling tty is good enough.
		   Under newer Unices, it requires an ioctl().  */
#ifdef TIOCSCTTY
		(void) ioctl(0, TIOCSCTTY, 0);
#endif

#ifdef SIGTSTP
		signal(SIGTSTP, SIG_IGN);
#endif
		
		(void) tty_reset(0);
		
		sprintf(str_addr,"KVADDR=%d",addr);
		putenv(str_addr);
		putenv("LINES=24");
		putenv("COLUMNS=80");
		putenv("TERM=kv");		
		putenv("TERMCAP=vt52|kvt|KeenView emulando dec vt52:\
			co#80:it#8:li#24:ae=\\EG:as=\\EF:bl=^G:cd=^J:\
			ce=\\EK:cl=\\EH\\EJ:cm:\\EY%+ %+ :cr=^M:do=\\EB:\
			ho=^L:kb=^H:kd=\\EB:kl=\\ED:kr=\\EC:ku=\\EA:le=\\ED:\
			nd=\\EC:nw=^M^J:sf=^J:sr=\\EI:ta=^I:up=\\EA:");
		putenv("PS1=%");
		sprintf(envline,"KV=%s",get_ttyname());
		putenv(envline);
		sprintf(envline,"KVADDR=%d",addr);
		putenv(envline);
		
		/* "touch" the tty so 'w' reports proper idle times */
		(void) utime(get_ttyname(), NULL);

		/* Set our uid to our real uid if necessary */
		(void) setuid(getuid());
		/*setuid(405);*/
		
		tty_sane(0);
		
		/*printf("\f==KeenView== %02d\n",addr);*/
		/* Run the requested program, with possible leading dash. */
	 	execvp(((*argv[0] == '-') ? argv[0]+1 : argv[0]), argv);

		perror(argv[0]);
		exit(255);
	}
	return(returnfd);
}



/*
 * Pseudo-terminal routines for Unix System V Release 3.2 and BSD4.2-3
 */


extern int errno;
int master_fd;

char tty_name[64]={'\0'};
char pty_name[64]={'\0'};

char *get_ttyname()
{
	if ( tty_name[0] )
		return(tty_name);
	return(NULL);
}

#include	<setjmp.h>

#ifndef X_OK
#define	R_OK	4	/* Test for Read permission */
#define	W_OK	2	/* Test for Write permission */
#define	X_OK	1	/* Test for eXecute permission */
#endif  /* X_OK */

jmp_buf next;

void trynext()
{
	longjmp(next, 2);
}


int get_master_pty()
{
	int i, master_fd;
	char *ptr;
	struct stat statbuff;
	static char ptychar[]=PTYCHAR;			/* X */ 
	static char hexdigit[]=HEXDIGIT;		/* Y */

	for (ptr=ptychar; *ptr != 0; ptr++)
	{
		strcpy(pty_name, "/dev/ptyXY");
		pty_name[8]=(*ptr);  /* X */
		pty_name[9]='0';   /* Y */

		if ( stat(pty_name, &statbuff) < 0 )
			break;
		i=(-1);		/* Initialize i */

		/* Set a time limit for the open */
		if ( setjmp(next) == -1 )
			return(-1);
		signal(SIGALRM, trynext);

		for ( ++i; hexdigit[i]; ++i)
		{
			pty_name[5]='p';
			pty_name[9]=hexdigit[i];

			alarm(2);	/* Set an open timeout */

			if ( (master_fd=open(pty_name, O_RDWR)) >= 0 )
			{
				alarm(0); 	/* Reset the alarm */

				pty_name[5]='t';
				sprintf(tty_name, "%s", pty_name);
				if ( access(tty_name, R_OK|W_OK) == 0 ) {
					signal(SIGALRM, SIG_DFL);
					return (master_fd);
				} else {
					pty_name[5]='p';
					(void) close(master_fd);
				}
			}
			/* reset the alarm */
			alarm(0);
		}
	}
	return(-1);
}


/* Open the slave half of a pseudo-terminal. */

int get_slave_pty()
{
	int slave_fd;

	if ( (slave_fd=open(tty_name, O_RDWR)) < 0 )
	{
		close(master_fd);
		return(-1);
	}
	return(slave_fd);
}


void d_zero(dst, len)
    register char *dst;
    register int len;
    {
	while (--len >= 0) *dst++ = 0;
    }

/* Here are the Terminal manipulation routines...  */

/* Get the modes of the controlling tty and save them.  Saves
   ttymodes in tty_mode and returns -1 if ioctl fails. */

struct termios tty_mode;  /* Save tty mode here */
static int tty_init=0;

int tty_getmode(fd)
int fd;
{
	d_zero((char *)&tty_mode, sizeof(struct termios));
	tty_init=1;	/* Flag: we have initialized the tty_mode struct */

	if ( ! isatty(fd) )
		return(0);

#ifdef OLDDEBUG
	fprintf(stderr, "Getting tty modes for tty_mode.\r\n");
#endif

	if (ioctl(fd, TCGETA, (char *) &tty_mode) < 0)
	{
#ifdef DEBUG
		perror("tty_getmode(): ioctl error");
#endif
		return(-1);  
	}

	return(0);
}


/* Set a tty to a sane mode */

int tty_sane(fd)
int fd;
{
	struct termios temp_mode;
	struct winsize win;

	win.ws_row = 24;
	win.ws_col = 80;
	ioctl(0, TIOCSWINSZ, &win); 

	tcgetattr(fd,&temp_mode);
	
	temp_mode.c_iflag |= (BRKINT | IGNPAR | IXON);
	temp_mode.c_iflag &= ~(ICRNL | INLCR | IGNCR);

	temp_mode.c_oflag |= (OPOST | ONLRET);
	temp_mode.c_oflag &= ~(OCRNL | ONLCR | ONOCR);
	
	
	temp_mode.c_cflag |= (CS8 | CREAD);
	temp_mode.c_lflag |= (ISIG | ICANON | ECHO | ECHOE | ECHOK);

	temp_mode.c_cc[VERASE]=('?'^64);
	temp_mode.c_cc[VINTR]=('C'^64);
	temp_mode.c_cc[VEOF]=('D'^64);
	temp_mode.c_cc[VQUIT]=('\\'^64);
	temp_mode.c_cc[VSTART]=('Q'^64);
	temp_mode.c_cc[VSTOP]=('S'^64);
	temp_mode.c_cc[VSUSP]=('Z'^64);
	temp_mode.c_cc[VWERASE]=('W'^64);
	temp_mode.c_cc[VDISCARD]=('O'^64);
	temp_mode.c_cc[VREPRINT]=('R'^64);
	temp_mode.c_cc[VLNEXT]=('V'^64);	
		
	tcsetattr(fd,TCSANOW,&temp_mode);

	
	return(0);
}

/* Restore terminal's mode to whatever it was on the most
   recent call to the tty_getmode() function. */

int tty_reset(fd)
int fd;
{
	if ( ! tty_init )
		return(-1);
	if ( ! isatty(fd) )
		return(0);
	/* TCSETAW is important for letting tty input drain. */
	if (ioctl(fd, TCSETAW, (char *) &tty_mode) < 0)
		return(-1);
	return(0);
}

int set_nonblock (int fd, int value) {
	int oldflags = fcntl(fd, F_GETFL,0);
	if (oldflags < 0) return oldflags;
	if (value)
		oldflags |= O_NONBLOCK;
	else 
		oldflags &= ~O_NONBLOCK;
	return fcntl(fd, F_SETFL, oldflags);
}
